""" Module providing disovery method of tap-dixa"""
import json
from singer import metadata
from singer.catalog import Catalog
from tap_dixa.streams import STREAMS, ActivityLogs
from tap_dixa.client import Client
from tap_dixa.helpers import (
    _get_key_properties_from_meta,
    _get_replication_key_from_meta,
    _get_replication_method_from_meta,
    get_abs_path,
    DixaURL
)
from datetime import datetime

def get_schemas():
    """
    Builds the singer schema and metadata dictionaries.
    """

    schemas = {}
    schemas_metadata = {}

    for stream_name, stream_object in STREAMS.items():

        schema_path = get_abs_path(f"schemas/{stream_name}.json")
        with open(schema_path) as file:
            schema = json.load(file)

        if stream_object.replication_method == "INCREMENTAL":
            replication_keys = stream_object.valid_replication_keys
        else:
            replication_keys = None

        meta = metadata.get_standard_metadata(schema=schema,
                                              key_properties=stream_object.key_properties,
                                              replication_method=stream_object.replication_method,
                                              valid_replication_keys=replication_keys,)

        meta = metadata.to_map(meta)

        if replication_keys:
            for replication_key in replication_keys:
                meta = metadata.write(meta,
                                      ("properties", replication_key),
                                      "inclusion",
                                      "automatic")

        meta = metadata.to_list(meta)

        schemas[stream_name] = schema
        schemas_metadata[stream_name] = meta

    return schemas, schemas_metadata


def discover(config: dict):
    """
    Builds the singer catalog for all the streams in the schemas directory.
    """

    schemas, schemas_metadata = get_schemas()
    streams = []

    if config:
        #Token Validation check before making any api request
        #params : mock parameter values are given for api token validation
        Client(config["api_token"]).get(base_url=DixaURL.INTEGRATIONS.value,
                                        endpoint=ActivityLogs.endpoint,
                                        params={"created_after": datetime.today(),
                                        "created_before": datetime.now()})

    for schema_name, schema in schemas.items():
        schema_meta = schemas_metadata[schema_name]

        catalog_entry = {
            "stream": schema_name,
            "tap_stream_id": schema_name,
            "schema": schema,
            "key_properties": _get_key_properties_from_meta(schema_meta),
            "replication_method": _get_replication_method_from_meta(schema_meta),
            "replication_key": _get_replication_key_from_meta(schema_meta),
            "metadata": schema_meta,
        }

        streams.append(catalog_entry)

    return Catalog.from_dict({"streams": streams})



import datetime


import singer
from tap_dixa.helpers import datetime_to_unix_ms,unix_ms_to_date_utc,DixaURL
from .abstracts import IncrementalStream


class Conversations(IncrementalStream):
    """
    Get conversations from the Dixa platform.
    """

    tap_stream_id = "conversations"
    key_properties = ["id"]
    replication_key = "updated_at"
    valid_replication_keys = ["updated_at"]
    old_replication_key = "updated_at_datestring"
    base_url = DixaURL.EXPORTS.value
    endpoint = "/v1/conversation_export"


    # pylint: disable=signature-differs
    def get_records(self, start_date :int):
        add_interval = datetime.timedelta(hours=self.get_interval())
        updated_after = unix_ms_to_date_utc(start_date)
        end_dt,loop = singer.utils.now(),True

        while loop:
            if (updated_after + add_interval) < end_dt:
                updated_before = updated_after + add_interval
            else:
                updated_before,loop = end_dt,False

            params = {"updated_after": datetime_to_unix_ms(updated_after),"updated_before": datetime_to_unix_ms(updated_before)}
            response = self.client.get(self.base_url, self.endpoint, params=params)

            yield from response

            updated_after = updated_before + datetime.timedelta(milliseconds=1)



import datetime

from tap_dixa.helpers import date_to_rfc3339, get_next_page_key, DixaURL
from .abstracts import IncrementalStream
import singer
from singer import metrics, Transformer


class ActivityLogs(IncrementalStream):
    """
    Get activity logs from the Dixa platform.
    """

    tap_stream_id = "activity_logs"
    key_properties = ["id"]
    replication_key = "activityTimestamp"
    valid_replication_keys = ["activityTimestamp"]
    base_url = DixaURL.INTEGRATIONS.value
    endpoint = "/v1/conversations/activitylog"

    def sync(self, state: dict, stream_schema: dict, stream_metadata: dict, config: dict, transformer: Transformer) -> dict:
        """
        The sync logic for an incremental stream.

        :param state: A dictionary representing singer state
        :param stream_schema: A dictionary containing the stream schema
        :param stream_metadata: A dictionnary containing stream metadata
        :param config: A dictionary containing tap config data
        :param transformer: A singer Transformer object
        :return: State data in the form of a dictionary
        """
        if config.get("interval"):
            self.set_interval(config.get("interval"))
        start_date = singer.get_bookmark(
            state, self.tap_stream_id, self.replication_key, config["start_date"])
        bookmark_datetime = singer.utils.strptime_to_utc(start_date)
        max_datetime = bookmark_datetime

        with metrics.record_counter(self.tap_stream_id) as counter:
            for record in self.get_records(bookmark_datetime, config=config):
                transformed_record = transformer.transform(record, stream_schema, stream_metadata)
                record_datetime = singer.utils.strptime_to_utc(transformed_record[self.replication_key])
                if record_datetime >= bookmark_datetime:
                    singer.write_record(self.tap_stream_id, transformed_record)
                    counter.increment()
                    max_datetime = max(record_datetime, max_datetime)

            bookmark_date = singer.utils.strftime(max_datetime)

        state = singer.write_bookmark(
            state, self.tap_stream_id, self.replication_key, bookmark_date)
        singer.write_state(state)
        return state

    # pylint: disable=signature-differs
    def get_records(self, start_date, config: dict = {}):
        max_limit = config.get("page_size", 10_000)
        loop = True
        page_key = None
        from_datetime = date_to_rfc3339(start_date.isoformat())
        to_datetime = date_to_rfc3339(datetime.datetime.utcnow().isoformat())

        params = {
            "fromDatetime": from_datetime,
            "toDatetime": to_datetime,
            "pageKey": page_key,
            "pageLimit": max_limit,
        }

        while loop:

            response = self.client.get(self.base_url, self.endpoint, params=params)

            # Extract data and pageKey
            data = response.get("data", [])
            meta = response.get("meta") or {}
            next_page = meta.get("next")
            page_key = get_next_page_key(next_page)

            # Update params with pageKey
            params.update({"pageKey": page_key.get("pageKey")})

            # Change switch to exit while loop if pageKey returns None
            loop = True if page_key.get("pageKey") else False

            yield from data



import datetime

import singer

from tap_dixa.helpers import datetime_to_unix_ms,unix_ms_to_date_utc, DixaURL
from .abstracts import IncrementalStream


class Messages(IncrementalStream):
    """
    Get messages from the Dixa platform.
    """

    tap_stream_id = "messages"
    key_properties = ["id"]
    replication_key = "created_at"
    valid_replication_keys = ["created_at"]
    old_replication_key = "updated_at_datestring"
    base_url = DixaURL.EXPORTS.value
    endpoint = "/v1/message_export"

    # pylint: disable=signature-differs
    def get_records(self, start_date :int):
        add_interval = datetime.timedelta(hours=self.get_interval())
        created_after = unix_ms_to_date_utc(start_date)
        end_dt,loop = singer.utils.now(),True


        while loop:
            if (created_after + add_interval) < end_dt:
                created_before = created_after + add_interval
            else:
                created_before,loop = end_dt,False

            params = {"created_after": datetime_to_unix_ms(created_after),"created_before": datetime_to_unix_ms(created_before)}
            response = self.client.get(self.base_url, self.endpoint, params=params)

            yield from response

            created_after = created_before + datetime.timedelta(milliseconds=1)



""" helper methods required for tap-dixa"""
import datetime
import os
import pytz

from enum import Enum
from typing import Iterator
from urllib.parse import parse_qsl, urlparse

from singer import  utils

def unix_ms_to_date(timestamp_ms: int) -> str:
    """
    Converts unix timestamp in milliseconds to ISO 8601 date string.

    :param ms: unix timestamp in milliseconds
    :return: ISO 8601 date string
    """
    return datetime.datetime.fromtimestamp(timestamp_ms / 1000).replace(microsecond=0).isoformat()


def unix_ms_to_date_utc(timestamp_ms: int) -> datetime.datetime:
    """
    Converts unix timestamp in milliseconds to timezone aware datetime object.

    :param timestamp_ms: unix timestamp in milliseconds
    :return: datetime obj
    """
    return datetime.datetime.fromtimestamp(timestamp_ms / 1000).replace(microsecond=0).replace(tzinfo=pytz.UTC)

def datetime_to_unix_ms(datetime_obj: datetime.datetime) -> int:
    """
    Converts datetime object to unix timestamp in milliseconds

    :param datetime_obj: A datetime object to convert to unix timestamp
    :return: integer representing unix timestamp in milliseconds
    """
    return int(datetime_obj.timestamp() * 1000)


def create_csid_params(csids: Iterator) -> dict:
    """
    Creates params for activity logs endpoint.

    Takes iterator and creates a comma-separated string list of the items
    in the iterator. Returns a dictionary with `csids` as key and
    comma-separated string list of conversation IDs as values.

    :param csids: Iterator containing conversation IDs
    :return: dictionary to be used with activity logs endpoint
    """
    return {"csids": ",".join(map(str, csids))}


def chunks(arr, chunk_size=10):
    """
    Takes an n-sized array and chunks it into smaller sizes that are
    yielded until full array has been iterated over.

    :param arr: n-sized list of conversation IDs
    :param chunk_size: the list size of each chunk
    :return: a chunk sized array
    """
    for i in range(0, len(arr), chunk_size):
        yield arr[i: i + chunk_size]


def date_to_rfc3339(date: str) -> str:
    """Converts date to rfc 3339"""
    date_utc = utils.strptime_to_utc(date)

    return date_utc.strftime("%Y-%m-%dT%H:%M:%SZ")


def get_next_page_key(url_part: str) -> dict:
    """
    Parses out and returns the `pageKey` value returned in the meta
    key from the API
    """
    parsed_url = urlparse(url_part)

    return dict(parse_qsl(parsed_url.query))


def get_abs_path(path):
    """
    Gets the absolute path of the provided relative path.
    """
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


def _get_key_properties_from_meta(schema_meta):
    """
    Retrieves the 'table-key-properties' value from the schema metadata.
    """
    return schema_meta[0].get("metadata").get("table-key-properties")


def _get_replication_method_from_meta(schema_meta):
    """
    Retrieves the 'forced-replication-method' value from the schema metadata.
    """
    return schema_meta[0].get("metadata").get("forced-replication-method")


def _get_replication_key_from_meta(schema_meta):
    """
    Retrieves the 'valid-replication-keys' value from the schema metadata.
    """
    if _get_replication_method_from_meta(schema_meta) == "INCREMENTAL":
        return schema_meta[0].get("metadata").get("valid-replication-keys")[0]
    return None


class Interval(Enum):
    """
    Enum representing time interval for making API calls.
    """

    HOUR = 1
    DAY = 24
    WEEK = 24 * 7
    MONTH = 24 * 31


class DixaURL(Enum):
    """
    Enum representing the Dixa base url API variants.
    """

    EXPORTS = "https://exports.dixa.io"
    INTEGRATIONS = "https://dev.dixa.io"



""" exceptions specific to dixa-client"""
import requests

from singer import get_logger

LOGGER = get_logger()


class InvalidInterval(Exception):
    pass


class DixaClientError(Exception):
    def __init__(self, message=None, response=None):
        super().__init__(message)
        self.message = message
        self.response = response


class DixaClient400Error(DixaClientError):
    pass

class DixaClient401Error(DixaClientError):
    pass

class DixaClient408Error(DixaClientError):
    pass

class DixaClient429Error(DixaClientError):
    pass


class DixaClient422Error(DixaClientError):
    pass

class DixaClient5xxError(DixaClientError):
    pass

ERROR_CODE_EXCEPTION_MAPPING = {
    400: {"raise_exception": DixaClient400Error, "message": "Invalid query parameters"},
    401: {"raise_exception": DixaClient401Error, "message": "Invalid or missing credentials"},
    408: {"raise_exception": DixaClient408Error, "message": "Request Timeout"},
    422: {"raise_exception": DixaClient422Error, "message": "Exceeded max allowed 10 csids per request"},
    429: {"raise_exception": DixaClient429Error, "message": "API limit has been reached"},
    500: {"raise_exception": DixaClient5xxError,"message": "Dixa Server Error",},
    503: {"raise_exception": DixaClient5xxError,"message": "Dixa Server Unavailable",},
}


def raise_for_error(resp: requests.Response):
    """
    Raises the associated response exception.

    Takes in a response object, checks the status code, and throws the associated
    exception based on the status code.

    :param resp: requests.Response object
    """
    try:
        resp.raise_for_status()
    except (requests.HTTPError, requests.ConnectionError) as error:
        try:
            error_code = resp.status_code
            client_exception = ERROR_CODE_EXCEPTION_MAPPING.get(error_code, {})
            exc = client_exception.get("raise_exception", DixaClientError)
            message = client_exception.get("message", "Client Error")

            raise exc(message, resp) from None

        except (ValueError, TypeError):
            raise DixaClientError(error) from None


def retry_after_wait_gen():
    """
    Returns a generator that is passed to backoff decorator to indicate how long
    to backoff for in seconds.
    """
    while True:
        sleep_time = 60
        LOGGER.info("API rate limit exceeded -- sleeping for %s seconds", sleep_time)
        yield sleep_time



from .activitylogs import ActivityLogs
from .conversations import Conversations
from .messages import Messages

STREAMS = {
    "activity_logs": ActivityLogs,
    "conversations": Conversations,
    "messages": Messages,
}


import singer
from singer import utils
from tap_dixa.discover import discover
from tap_dixa.sync import sync

REQUIRED_CONFIG_KEYS = ["start_date", "api_token"]
LOGGER = singer.get_logger()


@utils.handle_top_exception(LOGGER)
def main():
    # Parse command line arguments
    args = utils.parse_args(REQUIRED_CONFIG_KEYS)

    # If discover flag was passed, run discovery mode and dump output to stdout
    if args.discover:
        catalog = discover(args.config)
        catalog.dump()
    # Otherwise run in sync mode
    else:
        if args.catalog:
            catalog = args.catalog
        else:
            catalog = discover(args.config)
        sync(args.config, args.state, catalog)


if __name__ == "__main__":
    main()



import datetime
from abc import ABC, abstractmethod

import singer
from tap_dixa.client import Client
from tap_dixa.exceptions import InvalidInterval
from tap_dixa.helpers import Interval,datetime_to_unix_ms

LOGGER = singer.get_logger()


class BaseStream(ABC):
    """
    A base class representing singer streams.

    :param client: The API client used extract records from the external source
    """

    tap_stream_id = None
    replication_method = None
    replication_key = None
    key_properties = []
    valid_replication_keys = []
    params = {}
    endpoint = None
    base_url = None

    def __init__(self, client: Client):
        self.client = client

    @abstractmethod
    def get_records(self, start_date: datetime.datetime = None, config: dict = {}) -> list:
        """
        Returns a list of records for that stream.

        :param start_date: The start date datetime object the stream should use
        :param is_parent: If true, may change the type of data
            that is returned for a child stream to consume
        :return: list of records
        """
        raise NotImplementedError("Child classes of BaseStream require implementation")

    def set_parameters(self, params: dict) -> None:
        """
        Sets or updates the `params` attribute of a class.

        :param params: Dictionary of parameters to set or update the class with
        """
        self.params = params


class IncrementalStream(BaseStream):
    """
    A child class of a base stream used to represent streams that use the
    INCREMENTAL replication method.

    :param client: The API client used extract records from the external source
    """

    replication_method = "INCREMENTAL"
    batched = False
    interval = None
    old_replication_key = None
    
    def get_bookmark(self,state :dict,config: dict) ->int:
        """
        A wrapper for singer.get_bookmark to deal with backward compatibility for bookmark values.
        :param state: A dictionary representing singer state
        :param config: A dictionary containing tap config data
        :return: epoch timestamp in the form of a int datatype
        """
        bookmark = singer.get_bookmark(state, self.tap_stream_id, self.replication_key, False)
        if not bookmark:
            # get previous bookmark value if the current dosent exists or default to start date
            _ = singer.get_bookmark(state, self.tap_stream_id,self.old_replication_key, config["start_date"])   
            return datetime_to_unix_ms(singer.utils.strptime_to_utc(_))
        return bookmark

    def set_interval(self, value):
        """
        Sets the interval attribute.

        :param value: The interval string value
        """
        self.interval = value.upper()

    def get_interval(self):
        """
        Retrieves interval enum. Defaults to MONTH if no interval provided.
        :return: An enum for the interval to be used with the API
        """
        if self.interval:
            if hasattr(Interval, self.interval):
                return getattr(Interval, self.interval).value

            # If interval not part of enum, log message and throw error
            valid_intervals = set()
            for interval in dir(Interval):
                if not interval.startswith("__"):
                    valid_intervals.add(interval)

            # pylint: disable=logging-fstring-interpolation
            LOGGER.critical(f"provided interval '{self.interval}' is not " f"in Interval set: {valid_intervals}")

            raise InvalidInterval("invalid interval provided")

        return Interval.MONTH.value

    def sync(self, state: dict, stream_schema: dict, stream_metadata: dict, config: dict, transformer: singer.Transformer) -> dict:
        """
        The sync logic for an incremental stream.

        :param state: A dictionary representing singer state
        :param stream_schema: A dictionary containing the stream schema
        :param stream_metadata: A dictionnary containing stream metadata
        :param config: A dictionary containing tap config data
        :param transformer: A singer Transformer object
        :return: State data in the form of a dictionary
        """
        if config.get("interval"):
            self.set_interval(config.get("interval"))
        start_date_epoch = self.get_bookmark(state,config)
        # bookmark_datetime = singer.utils.strptime_to_utc(start_date)
        max_datetime = bookmark_datetime = start_date_epoch

        with singer.metrics.record_counter(self.tap_stream_id) as counter:
            for record in self.get_records(bookmark_datetime):
                transformed_record = transformer.transform(record, stream_schema, stream_metadata)
                record_datetime = transformed_record[self.replication_key]
                if record_datetime >= bookmark_datetime:
                    singer.write_record(self.tap_stream_id, transformed_record)
                    counter.increment()
                    max_datetime = max(record_datetime, max_datetime)

            bookmark_date = max_datetime

        state = singer.write_bookmark(state, self.tap_stream_id, self.replication_key, bookmark_date)
        singer.write_state(state)
        return state


class FullTableStream(BaseStream):
    """
    A child class of a base stream used to represent streams that use the
    FULL_TABLE replication method.

    :param client: The API client used extract records from the external source
    """

    replication_method = "FULL_TABLE"

    def sync(self, state: dict, stream_schema: dict, stream_metadata: dict, config: dict, transformer: singer.Transformer) -> dict:
        """
        The sync logic for an full table stream.

        :param state: A dictionary representing singer state
        :param stream_schema: A dictionary containing the stream schema
        :param stream_metadata: A dictionnary containing stream metadata
        :param config: A dictionary containing tap config data
        :param transformer: A singer Transformer object
        :return: State data in the form of a dictionary
        """
        with singer.metrics.record_counter(self.tap_stream_id) as counter:
            for record in self.get_records(config):
                transformed_record = transformer.transform(record, stream_schema, stream_metadata)
                singer.write_record(self.tap_stream_id, transformed_record)
                counter.increment()

        singer.write_state(state)
        return state



from setuptools import find_packages, setup

requirements_file = "requirements.txt"

setup(
    name="tap-dixa",
    version="1.0.2",
    description="Singer.io tap for extracting data from the Dixa API",
    author="Stitch",
    url="http://singer.io",
    classifiers=["Programming Language :: Python :: 3 :: Only"],
    py_modules=["tap_dixa"],
    install_requires=[
        "backoff==1.8.0",
        "certifi==2021.5.30",
        "charset-normalizer==2.0.4",
        "ciso8601==2.1.3",
        "idna==3.2",
        "jsonschema==2.6.0",
        "python-dateutil==2.8.2",
        "pytz==2018.4",
        "requests==2.26.0",
        "simplejson==3.11.1",
        "singer-python==5.12.1",
        "six==1.16.0",
        "urllib3==1.26.6",
    ],
    entry_points="""
    [console_scripts]
    tap-dixa=tap_dixa:main
    """,
    packages=find_packages(exclude=["tests"]),
    package_data = {
        "schemas": ["tap_dixa/schemas/*.json"]
    },
    include_package_data=True,
)



""" Module providing DixaAPi Client"""
import base64
import backoff
import requests

from tap_dixa.exceptions import (DixaClient429Error, DixaClient408Error, 
                                DixaClient5xxError, raise_for_error,
                                retry_after_wait_gen)
from tap_dixa.helpers import DixaURL

class Client:
    """DixaClient Class for performing extraction from DixaApi"""

    def __init__(self, api_token: str):
        self._api_token = api_token
        self._base_url = None
        self._session = requests.Session()
        self._headers = {}

    @staticmethod
    def _to_base64(string: str) -> str:
        """
        Base64 encodes a string.

        :param string: String to be base64 encoded
        :return: Base64 encoded string
        """
        message = f"bearer:{string}"
        message_bytes = message.encode("utf-8")
        base64_bytes = base64.b64encode(message_bytes)
        return base64_bytes.decode("utf-8")

    def _set_auth_header(self):
        """
        Sets the corresponding Authorization header based on the base url variant.
        """
        if self._base_url == DixaURL.EXPORTS.value:
            self._headers["Authorization"] = f"Basic {self._to_base64(self._api_token)}"
        elif self._base_url == DixaURL.INTEGRATIONS.value:
            self._headers["Authorization"] = f"{self._api_token}"

    def _build_url(self, endpoint: str) -> str:
        """
        Builds the URL for the API request.

        :param endpoint: The API URI (resource)
        :return: The full API URL for the request
        """
        return f"{self._base_url}{endpoint}"

    def _get(self, url, headers=None, params=None, data=None):
        """
        Wraps the _make_request function with a 'GET' method
        """
        return self._make_request(url, method="GET", headers=headers, params=params, data=data)

    def _post(self, url, headers=None, params=None, data=None):
        """
        Wraps the _make_request function with a 'POST' method
        """
        return self._make_request(url, method="POST", headers=headers, params=params, data=data)

    # Added retry logic for 3 times when bad request or server error or rate limit happens
    @backoff.on_exception(retry_after_wait_gen, (DixaClient429Error, DixaClient5xxError,DixaClient408Error), jitter=None, max_tries=3)
    def _make_request(self, url, method, headers=None, params=None, data=None) -> dict:
        """
        Makes the API request.

        :param url: The full API url
        :param method: The API request method
        :param headers: The headers for the API request
        :param params: The querystring params passed to the API
        :param data: The data passed to the body of the request
        :return: A dictionary representing the response from the API
        """
        with self._session as session:
            response = session.request(method, url, headers=headers, params=params, data=data)

            if response.status_code != 200:
                raise_for_error(response)
                return None

            return response.json()

    def get(self, base_url, endpoint, params=None):
        """
        Takes the base_url and endpoint and builds and makes a 'GET' request
        to the API.
        """
        self._base_url = base_url
        self._set_auth_header()
        url = self._build_url(endpoint)
        return self._get(url, headers=self._headers, params=params)



import singer
from singer import Transformer, metadata
from tap_dixa.client import Client
from tap_dixa.streams import STREAMS

LOGGER = singer.get_logger()


def sync(config, state, catalog):
    """Sync data from tap source"""

    client = Client(config.get("api_token"))

    with Transformer() as transformer:
        for stream in catalog.get_selected_streams(state):
            tap_stream_id = stream.tap_stream_id
            stream_obj = STREAMS[tap_stream_id](client)
            stream_schema = stream.schema.to_dict()
            stream_metadata = metadata.to_map(stream.metadata)

            LOGGER.info("Starting sync for stream: %s", tap_stream_id)

            state = singer.set_currently_syncing(state, tap_stream_id)
            singer.write_state(state)

            singer.write_schema(tap_stream_id, stream_schema, stream_obj.key_properties, stream.replication_key)

            state = stream_obj.sync(state, stream_schema, stream_metadata, config, transformer)
            singer.write_state(state)

    state = singer.set_currently_syncing(state, None)
    singer.write_state(state)
